@@ -9,4 +9,4 @@
 void dmenu_load_entries(
     struct application_list *applications, char delim, int event_fd, int abort_fd);
 bool dmenu_execute(const struct application *app, ssize_t index,
-                   const struct prompt *prompt, enum dmenu_mode format);
+                   const struct prompt *prompt, enum dmenu_mode format, unsigned int column);
