@@ -29,6 +29,8 @@
 #define LOG_ENABLE_DBG 0
 #include "log.h"
 #include "xdg.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 typedef tll(char *) theme_names_t;
 
@@ -95,7 +97,7 @@ parse_theme(FILE *index, struct icon_theme *theme, theme_names_t *themes_to_load
             type = ICON_DIR_THRESHOLD;
             threshold = 2;
 
-            section = malloc(len - 2 + 1);
+            section = xmalloc(len - 2 + 1);
             memcpy(section, &line[1], len - 2);
             section[len - 2] = '\0';
             free(line);
@@ -112,7 +114,7 @@ parse_theme(FILE *index, struct icon_theme *theme, theme_names_t *themes_to_load
             for (const char *theme_name = strtok_r(value, ",", &ctx);
                  theme_name != NULL; theme_name = strtok_r(NULL, ",", &ctx))
             {
-                tll_push_back(*themes_to_load, strdup(theme_name));
+                tll_push_back(*themes_to_load, xstrdup(theme_name));
             }
         }
 
@@ -122,7 +124,7 @@ parse_theme(FILE *index, struct icon_theme *theme, theme_names_t *themes_to_load
                  d != NULL;
                  d = strtok_r(NULL, ",", &save))
             {
-                struct icon_dir dir = {.path = strdup(d)};
+                struct icon_dir dir = {.path = xstrdup(d)};
                 tll_push_back(theme->dirs, dir);
             }
         }
@@ -140,7 +142,7 @@ parse_theme(FILE *index, struct icon_theme *theme, theme_names_t *themes_to_load
             sscanf(value, "%d", &scale);
 
         else if (strcasecmp(key, "context") == 0)
-            context = strdup(value);
+            context = xstrdup(value);
 
         else if (strcasecmp(key, "threshold") == 0)
             sscanf(value, "%d", &threshold);
@@ -192,9 +194,9 @@ load_theme_in(const char *dir, struct icon_theme *theme,
               theme_names_t *themes_to_load)
 {
     char path[PATH_MAX];
-    snprintf(path, sizeof(path), "%s/index.theme", dir);
+    xsnprintf(path, sizeof(path), "%s/index.theme", dir);
 
-    FILE *index = fopen(path, "r");
+    FILE *index = fopen(path, "re");
     if (index == NULL)
         return false;
 
@@ -227,12 +229,53 @@ discover_and_load_theme(const char *theme_name, xdg_data_dirs_t dirs,
 
         struct icon_theme theme = {0};
         if (load_theme_in(path, &theme, themes_to_load)) {
-            theme.name = strdup(theme_name);
+            theme.name = xstrdup(theme_name);
             tll_push_back(*themes, theme);
         }
     }
 }
 
+static xdg_data_dirs_t
+get_icon_dirs(void)
+{
+    /*
+     * See https://specifications.freedesktop.org/icon-theme-spec/latest/#directory_layout
+     */
+
+    xdg_data_dirs_t dirs = xdg_data_dirs();
+
+    const char *home = getenv("HOME");
+    if (home != NULL) {
+        char *path = xasprintf("%s/%s", home, ".icons");
+        int fd = open(path, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+
+        if (fd >= 0) {
+            struct xdg_data_dir home_icons = {
+                .path = path,
+                .fd = fd,
+            };
+
+            tll_push_front(dirs, home_icons);
+        } else
+            free(path);
+    }
+
+    {
+        const char *path = "/usr/share/pixmaps";
+        int fd = open(path, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+        if (fd >= 0) {
+            struct xdg_data_dir home_icons = {
+                .path = xstrdup(path),
+                .fd = fd,
+            };
+
+            tll_push_back(dirs, home_icons);
+        }
+    }
+
+    return dirs;
+}
+
 icon_theme_list_t
 icon_load_theme(const char *name)
 {
@@ -243,9 +286,9 @@ icon_load_theme(const char *name)
     /* List of themes to try to load. This list will be appended to as
      * we go, and find 'Inherits' values in the theme index files. */
     theme_names_t themes_to_load = tll_init();
-    tll_push_back(themes_to_load, strdup(name));
+    tll_push_back(themes_to_load, xstrdup(name));
 
-    xdg_data_dirs_t dirs = xdg_data_dirs();
+    xdg_data_dirs_t dirs = get_icon_dirs();
 
     while (tll_length(themes_to_load) > 0) {
         char *theme_name = tll_pop_front(themes_to_load);
@@ -379,7 +422,7 @@ icon_from_svg(struct icon *icon, const char *name)
 static bool
 svg(struct icon *icon, const char *path)
 {
-    icon->path = strdup(path);
+    icon->path = xstrdup(path);
     icon->type = ICON_SVG;
     icon->svg = NULL;
     return true;
@@ -388,7 +431,7 @@ svg(struct icon *icon, const char *path)
 static bool
 png(struct icon *icon, const char *path)
 {
-    icon->path = strdup(path);
+    icon->path = xstrdup(path);
     icon->type = ICON_PNG;
     icon->png = NULL;
     return true;
@@ -529,16 +572,12 @@ lookup_icons(const icon_theme_list_t *themes, int icon_size,
                     LOG_DBG("%s: abslute path PNG", app->icon.name);
             }
         } else {
-            size_t file_name_len = strlen(app->icon.name) + 4;
-            char *file_name = malloc(file_name_len + 1);
-            strcpy(file_name, app->icon.name);
-            strcat(file_name, ".xxx");
-
+            char *file_name = xstrjoin(app->icon.name, ".xxx");
             struct icon_data data = {
                 .name = app->icon.name,
                 .app = app,
                 .file_name = file_name,
-                .file_name_len = file_name_len,
+                .file_name_len = strlen(file_name),
                 .min_diff = {.diff = INT_MAX},
             };
             tll_push_back(icons, data);
@@ -567,11 +606,21 @@ lookup_icons(const icon_theme_list_t *themes, int icon_size,
             tll_foreach(*xdg_dirs, xdg_dir_it) {
                 const struct xdg_data_dir *xdg_dir = &xdg_dir_it->item;
 
-                const int scale = icon_dir->scale;
-                const int size = icon_dir->size * scale;
-                const int min_size = icon_dir->min_size * scale;
-                const int max_size = icon_dir->max_size * scale;
-                const int threshold = icon_dir->threshold * scale;
+                if (icon_dir->scale > 1) {
+                    /*
+                     * Scaled dirs are physically bigger icons (more
+                     * pixels), but with less details. See
+                     *  - https://codeberg.org/dnkl/fuzzel/issues/459#issuecomment-2574718
+                     *  - https://codeberg.org/dnkl/fuzzel/issues/459#issuecomment-2574720
+                     * For details on why we're skipping these.
+                     */
+                    continue;
+                }
+
+                const int size = icon_dir->size;
+                const int min_size = icon_dir->min_size;
+                const int max_size = icon_dir->max_size;
+                const int threshold = icon_dir->threshold;
                 const enum icon_dir_type type = icon_dir->type;
 
                 bool is_exact_match = false;
@@ -622,7 +671,7 @@ lookup_icons(const icon_theme_list_t *themes, int icon_size,
                 }
 
                 int dir_fd = openat(
-                    xdg_dir->fd, theme_relative_path, O_RDONLY | O_DIRECTORY);
+                    xdg_dir->fd, theme_relative_path, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
                 if (dir_fd < 0)
                     continue;
 
@@ -650,15 +699,9 @@ lookup_icons(const icon_theme_list_t *themes, int icon_size,
                         continue;
                     }
 
-                    char *full_path = malloc(
-                        strlen(xdg_dir->path) + 1 +
-                        5 + 1 + /* “icons” */
-                        strlen(theme->name) + 1 +
-                        strlen(icon_dir->path) + 1 +
-                        len + 1);
-
-                    sprintf(full_path, "%s/icons/%s/%s/%s",
-                            xdg_dir->path, theme->name, icon_dir->path, path);
+                    char *full_path = xasprintf(
+                        "%s/icons/%s/%s/%s",
+                        xdg_dir->path, theme->name, icon_dir->path, path);
 
                     if ((path[len - 3] == 's' &&
                          svg(&icon->app->icon, full_path)) ||
@@ -726,34 +769,24 @@ lookup_icons(const icon_theme_list_t *themes, int icon_size,
         const struct icon_data *icon = &icon_it->item;
 
         tll_foreach(*xdg_dirs, it) {
-            int pixmaps_fd = openat(it->item.fd, "pixmaps", O_RDONLY);
-            if (pixmaps_fd < 0)
-                continue;
-
             size_t len = icon->file_name_len;
             char *path = icon->file_name;
 
-            if (!icon_file_exists(pixmaps_fd, path, len)) {
-                close(pixmaps_fd);
+            if (!icon_file_exists(it->item.fd, path, len))
                 continue;
-            }
 
-            close(pixmaps_fd);
-
-            char full_path[strlen(it->item.path) + 1 +
-                           strlen("pixmaps") + 1 +
-                           len + 1];
+            char full_path[strlen(it->item.path) + 1 + len + 1];
 
             /* Try SVG variant first */
-            sprintf(full_path, "%s/pixmaps/%s", it->item.path, path);
+            sprintf(full_path, "%s/%s", it->item.path, path);
             if (path[len - 3] == 's' && svg(&icon->app->icon, full_path)) {
-                LOG_DBG("%s: %s (pixmaps)", icon->name, full_path);
+                LOG_DBG("%s: %s (standalone)", icon->name, full_path);
                 break;
             }
 
             /* No SVG, look for PNG instead */
             if (path[len - 3] == 'p' && png(&icon->app->icon, full_path)) {
-                LOG_DBG("%s: %s (pixmaps)", icon->name, full_path);
+                LOG_DBG("%s: %s (standalone)", icon->name, full_path);
                 break;
             }
 
@@ -770,7 +803,7 @@ bool
 icon_lookup_application_icons(icon_theme_list_t themes, int icon_size,
                               struct application_list *applications)
 {
-    xdg_data_dirs_t xdg_dirs = xdg_data_dirs();
+    xdg_data_dirs_t xdg_dirs = get_icon_dirs();
     lookup_icons(&themes, icon_size, applications, &xdg_dirs);
     xdg_data_dirs_destroy(xdg_dirs);
 
