@@ -38,6 +38,8 @@
 #include "version.h"
 #include "wayland.h"
 #include "xdg.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 #define max(x, y) ((x) > (y) ? (x) : (y))
 
@@ -54,6 +56,7 @@ struct context {
     mtx_t *icon_lock;
 
     const char *select_initial;
+    const size_t select_initial_idx;
 
     int event_fd;
     int dmenu_abort_fd;
@@ -83,7 +86,7 @@ read_cache(const char *path, struct application_list *apps, bool dmenu)
             return;
         }
 
-        int cache_dir_fd = open(path, O_DIRECTORY);
+        int cache_dir_fd = open(path, O_DIRECTORY | O_CLOEXEC);
         if (cache_dir_fd == -1) {
             LOG_ERRNO("%s: failed to open", path);
             return;
@@ -91,7 +94,7 @@ read_cache(const char *path, struct application_list *apps, bool dmenu)
 
         int fd = -1;
         if (fstatat(cache_dir_fd, "fuzzel", &st, 0) < 0 ||
-            (fd = openat(cache_dir_fd, "fuzzel", O_RDONLY)) < 0 ||
+            (fd = openat(cache_dir_fd, "fuzzel", O_RDONLY | O_CLOEXEC)) < 0 ||
             (f = fdopen(fd, "r")) == NULL)
         {
             close(cache_dir_fd);
@@ -104,7 +107,7 @@ read_cache(const char *path, struct application_list *apps, bool dmenu)
         }
         close(cache_dir_fd);
     } else {
-        if (stat(path, &st) < 0 || (f = fopen(path, "r")) == NULL)
+        if (stat(path, &st) < 0 || (f = fopen(path, "re")) == NULL)
         {
             if (errno != ENOENT)
                 LOG_ERRNO("%s: failed to open", path);
@@ -150,7 +153,7 @@ read_cache(const char *path, struct application_list *apps, bool dmenu)
         sscanf(count_str, "%u", &count);
 
         struct cache_entry entry = {
-            .id = dmenu ? NULL : strdup(id),
+            .id = dmenu ? NULL : xstrdup(id),
             .title = dmenu ? ambstoc32(id) : NULL,
             .count = count,
         };
@@ -210,16 +213,16 @@ write_cache(const char *path, const struct application_list *apps, bool dmenu)
             return;
         }
 
-        int cache_dir_fd = open(path, O_DIRECTORY);
+        int cache_dir_fd = open(path, O_DIRECTORY | O_CLOEXEC);
         if (cache_dir_fd == -1) {
             LOG_ERRNO("%s: failed to open", path);
             return;
         }
 
-        fd = openat(cache_dir_fd, "fuzzel", O_WRONLY | O_CREAT | O_TRUNC, 0644);
+        fd = openat(cache_dir_fd, "fuzzel", O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0644);
         close(cache_dir_fd);
     } else {
-        fd = open(path, O_WRONLY | O_CREAT | O_TRUNC, 0644);
+        fd = open(path, O_WRONLY | O_CREAT | O_TRUNC | O_CLOEXEC, 0644);
     }
 
     if (fd == -1) {
@@ -264,8 +267,8 @@ write_cache(const char *path, const struct application_list *apps, bool dmenu)
         }
 
         char count_as_str[11];
-        sprintf(count_as_str, "%u", apps->v[i]->count);
-        const size_t count_len = strlen(count_as_str);
+        const size_t count_len = xsnprintf(
+            count_as_str, sizeof(count_as_str), "%u", apps->v[i]->count);
 
         if (write(fd, "|", 1) != 1 ||
             write(fd, count_as_str, count_len) != count_len ||
@@ -283,7 +286,7 @@ static const char *
 version_and_features(void)
 {
     static char buf[256];
-    snprintf(buf, sizeof(buf), "version: %s %ccairo %cpng %csvg%s %cassertions",
+    xsnprintf(buf, sizeof(buf), "version: %s %ccairo %cpng %csvg%s %cassertions",
              FUZZEL_VERSION,
 #if defined(FUZZEL_ENABLE_CAIRO)
              '+',
@@ -347,6 +350,7 @@ print_usage(const char *prog_name)
            "     --y-margin=MARGIN           vertical margin, in pixels (0)\n"
            "     --select=STRING             select first entry that matches the given\n"
            "                                 string\n"
+           "     --select-index=INDEX        select the entry with index, not compatible with --select\n"
            "  -l,--lines                     number of matches to show\n"
            "  -w,--width                     window width, in characters (margins and\n"
            "                                 borders not included)\n"
@@ -385,6 +389,7 @@ print_usage(const char *prog_name)
            "     --letter-spacing=AMOUNT     additional letter spacing\n"
            "     --layer=top|overlay         which layer to render the fuzzel window on\n"
            "                                 (top)\n"
+           "     --keyboard-focus=exclusive|on-demand  keyboard focus mode (exclusive)\n"
            "     --no-exit-on-keyboard-focus-loss  do not exit when losing keyboard focus\n"
            "     --launch-prefix=COMMAND     prefix to add before argv of executed program\n"
            "     --list-executables-in-path  include executables from PATH in the list\n"
@@ -401,6 +406,10 @@ print_usage(const char *prog_name)
            "                                 instead of newline separated\n"
            "     --index                     print selected entry's index instead of of the \n"
            "                                 entry's text (dmenu mode only)\n"
+           "     --with-nth=N                display the N:th column (tab separated) of each\n"
+           "                                 input line (dmenu only)\n"
+           "     --accept-nth=N              output the N:th column (tab separated) of each\n"
+           "                                 input line (dmenu only)\n"
            "  -R,--no-run-if-empty           exit immediately without showing UI if stdin\n"
            "                                 is empty (dmenu mode only)\n"
            "     --log-level={info|warning|error|none}\n"
@@ -478,13 +487,7 @@ lock_file_name(void)
     if (wayland_display == NULL)
         return NULL;
 
-    #define path_fmt "%s/fuzzel-%s.lock"
-    int chars = snprintf(NULL, 0, path_fmt, xdg_runtime_dir, wayland_display);
-
-    char *path = malloc(chars + 1);
-    snprintf(path, chars + 1, path_fmt, xdg_runtime_dir, wayland_display);
-    #undef path_fmt
-
+    char *path = xasprintf("%s/fuzzel-%s.lock", xdg_runtime_dir, wayland_display);
     LOG_DBG("lock file: %s", path);
     return path;
 }
@@ -495,7 +498,7 @@ acquire_file_lock(const char *path, int *fd)
     *fd = open(path, O_WRONLY | O_CREAT | O_CLOEXEC, 0644);
     if (*fd < 0) {
         /* Warn, but allow running anyway */
-        LOG_WARN("%s: failed to create lock file", path);
+        LOG_WARN("%s: failed to create lock file: %s", path, strerror(errno));
         return true;
     }
 
@@ -532,12 +535,12 @@ populate_apps(void *_ctx)
     if (filter_desktop) {
         char *xdg_current_desktop = getenv("XDG_CURRENT_DESKTOP");
         if (xdg_current_desktop && strlen(xdg_current_desktop) != 0) {
-            xdg_current_desktop = strdup(xdg_current_desktop);
+            xdg_current_desktop = xstrdup(xdg_current_desktop);
             for (char *desktop = strtok_r(xdg_current_desktop, ":", &saveptr);
                  desktop != NULL;
                  desktop = strtok_r(NULL, ":", &saveptr))
                 {
-                    tll_push_back(desktops, strdup(desktop));
+                    tll_push_back(desktops, xstrdup(desktop));
                 }
             free(xdg_current_desktop);
         }
@@ -593,6 +596,7 @@ process_event(struct context *ctx, enum event_type event)
     struct application_list *apps = ctx->apps;
     struct matches *matches = ctx->matches;
     const char *select = ctx->select_initial;
+    const size_t select_idx = ctx->select_initial_idx;
 
     switch (event) {
     case EVENT_APPS_SOME_LOADED:
@@ -620,7 +624,14 @@ process_event(struct context *ctx, enum event_type event)
              *   c) a partial load will cause more matches to be displayed
              */
             matches_update_no_delay(matches);
-            matches_selected_select(matches, select);
+            if (select_idx != 0) {
+                if (!matches_selected_set(matches, select_idx)) {
+                    LOG_ERR("couldn't select entry at index %zu", select_idx);
+                    return false;
+                }
+            }
+            else
+                matches_selected_select(matches, select);
         }
         break;
     }
@@ -706,31 +717,35 @@ main(int argc, char *const *argv)
     #define OPT_CONFIG                       268
     #define OPT_LAYER                        269
     #define OPT_ICON_THEME                   270
-    #define OPT_NO_EXIT_ON_KB_LOSS           271
-    #define OPT_TABS                         272
-    #define OPT_DMENU_NULL                   273
-    #define OPT_FILTER_DESKTOP               274
-    #define OPT_CHECK_CONFIG                 275
-    #define OPT_SELECT                       276
-    #define OPT_LIST_EXECS_IN_PATH           277
-    #define OPT_X_MARGIN                     278
-    #define OPT_Y_MARGIN                     279
-    #define OPT_CACHE                        280
-    #define OPT_RENDER_WORKERS               281
-    #define OPT_PROMPT_COLOR                 282
-    #define OPT_INPUT_COLOR                  283
-    #define OPT_PROMPT_ONLY                  284
-    #define OPT_COUNTER_COLOR                285
-    #define OPT_USE_BOLD                     286
-    #define OPT_MATCH_WORKERS                287
-    #define OPT_NO_SORT                      288
-    #define OPT_DELAYED_FILTER_MS            289
-    #define OPT_DELAYED_FILTER_LIMIT         290
-    #define OPT_PLACEHOLDER                  291
-    #define OPT_PLACEHOLDER_COLOR            292
-    #define OPT_SEARCH_TEXT                  293
-    #define OPT_COUNTER                      294
-    #define OPT_HIDE_WHEN_PROMPT_EMPTY       295
+    #define OPT_KEYBOARD_FOCUS               271
+    #define OPT_NO_EXIT_ON_KB_LOSS           272
+    #define OPT_TABS                         273
+    #define OPT_DMENU_NULL                   274
+    #define OPT_FILTER_DESKTOP               275
+    #define OPT_CHECK_CONFIG                 276
+    #define OPT_SELECT                       277
+    #define OPT_SELECT_INDEX                 278
+    #define OPT_LIST_EXECS_IN_PATH           279
+    #define OPT_X_MARGIN                     280
+    #define OPT_Y_MARGIN                     281
+    #define OPT_CACHE                        282
+    #define OPT_RENDER_WORKERS               283
+    #define OPT_PROMPT_COLOR                 284
+    #define OPT_INPUT_COLOR                  285
+    #define OPT_PROMPT_ONLY                  286
+    #define OPT_COUNTER_COLOR                287
+    #define OPT_USE_BOLD                     288
+    #define OPT_MATCH_WORKERS                289
+    #define OPT_NO_SORT                      290
+    #define OPT_DELAYED_FILTER_MS            291
+    #define OPT_DELAYED_FILTER_LIMIT         292
+    #define OPT_PLACEHOLDER                  293
+    #define OPT_PLACEHOLDER_COLOR            294
+    #define OPT_SEARCH_TEXT                  295
+    #define OPT_COUNTER                      296
+    #define OPT_HIDE_WHEN_PROMPT_EMPTY       297
+    #define OPT_DMENU_WITH_NTH               298
+    #define OPT_DMENU_ACCEPT_NTH             299
 
     static const struct option longopts[] = {
         {"config",               required_argument, 0, OPT_CONFIG},
@@ -749,6 +764,7 @@ main(int argc, char *const *argv)
         {"x-margin",             required_argument, 0, OPT_X_MARGIN},
         {"y-margin",             required_argument, 0, OPT_Y_MARGIN},
         {"select",               required_argument, 0, OPT_SELECT},
+        {"select-index",         required_argument, 0, OPT_SELECT_INDEX},
         {"lines",                required_argument, 0, 'l'},
         {"width",                required_argument, 0, 'w'},
         {"tabs",                 required_argument, 0, OPT_TABS},
@@ -783,8 +799,9 @@ main(int argc, char *const *argv)
         {"letter-spacing",       required_argument, 0, OPT_LETTER_SPACING},
         {"launch-prefix",        required_argument, 0, OPT_LAUNCH_PREFIX},
         {"layer",                required_argument, 0, OPT_LAYER},
-        {"no-exit-on-keyboard-focus-loss", no_argument, 0, OPT_NO_EXIT_ON_KB_LOSS},
-        {"list-executables-in-path",       no_argument, 0, OPT_LIST_EXECS_IN_PATH},
+        {"keyboard-focus",       required_argument, 0, OPT_KEYBOARD_FOCUS},
+        {"no-exit-on-keyboard-focus-loss",   no_argument, 0, OPT_NO_EXIT_ON_KB_LOSS},
+        {"list-executables-in-path",         no_argument, 0, OPT_LIST_EXECS_IN_PATH},
         {"render-workers",       required_argument, 0, OPT_RENDER_WORKERS},
         {"match-workers",        required_argument, 0, OPT_MATCH_WORKERS},
         {"no-sort",              no_argument,       0, OPT_NO_SORT},
@@ -797,6 +814,8 @@ main(int argc, char *const *argv)
         {"dmenu0",               no_argument,       0, OPT_DMENU_NULL},
         {"no-run-if-empty",      no_argument,       0, 'R'},
         {"index",                no_argument,       0, OPT_DMENU_INDEX},
+        {"with-nth",             required_argument, 0, OPT_DMENU_WITH_NTH},
+        {"accept-nth",           required_argument, 0, OPT_DMENU_ACCEPT_NTH},
 
         /* Misc */
         {"log-level",            required_argument, 0, OPT_LOG_LEVEL},
@@ -813,6 +832,7 @@ main(int argc, char *const *argv)
     enum log_colorize log_colorize = LOG_COLORIZE_AUTO;
     bool log_syslog = true;
     const char *select = NULL;
+    size_t select_idx = 0;
 
     struct {
         struct config conf;
@@ -854,7 +874,10 @@ main(int argc, char *const *argv)
         bool dmenu_mode_set:1;
         bool dmenu_exit_immediately_if_empty_set:1;
         bool dmenu_delim_set:1;
+        bool dmenu_with_nth_set:1;
+        bool dmenu_accept_nth_set:1;
         bool layer_set:1;
+        bool keyboard_focus_set:1;
         bool no_exit_on_keyboard_focus_loss_set:1;
         bool render_workers_set:1;
         bool match_workers_set:1;
@@ -872,16 +895,14 @@ main(int argc, char *const *argv)
 
     /* Auto-enable dmenu mode if invoked through a ‘dmenu’ symlink */
     if (argv[0] != NULL) {
-        char *copy = strdup(argv[0]);
-        if (copy != NULL) {
-            const char *name = basename(copy);
-            if (name != NULL && strcmp(name, "dmenu") == 0) {
-                cmdline_overrides.conf.dmenu.enabled = true;
-                cmdline_overrides.dmenu_enabled_set = true;
-            }
-
-            free(copy);
+        char *copy = xstrdup(argv[0]);
+        const char *name = basename(copy);
+        if (name != NULL && strcmp(name, "dmenu") == 0) {
+            cmdline_overrides.conf.dmenu.enabled = true;
+            cmdline_overrides.dmenu_enabled_set = true;
         }
+
+        free(copy);
     }
 
     while (true) {
@@ -973,7 +994,7 @@ main(int argc, char *const *argv)
             {
                 enum match_fields field = 0;
 
-                for (size_t i = 0; i < sizeof(map) / sizeof(map[0]); i++) {
+                for (size_t i = 0; i < ALEN(map); i++) {
                     if (strcmp(f, map[i].name) == 0) {
                         field = map[i].value;
                         break;
@@ -984,9 +1005,9 @@ main(int argc, char *const *argv)
                     cmdline_overrides.conf.match_fields |= field;
                 else {
                     char valid_names[128] = {0};
-                    int idx = 0;
-                    for (size_t i = 0; i < sizeof(map) / sizeof(map[0]); i++) {
-                        idx += snprintf(
+                    size_t idx = 0;
+                    for (size_t i = 0; i < ALEN(map); i++) {
+                        idx += xsnprintf(
                             &valid_names[idx], sizeof(valid_names) - idx,
                             "'%s', ", map[i].name);
                     }
@@ -1078,6 +1099,16 @@ main(int argc, char *const *argv)
             select = optarg;
             break;
 
+        case OPT_SELECT_INDEX:
+            if (sscanf(optarg, "%zu", &select_idx) != 1) {
+                fprintf(
+                    stderr,
+                    "%s: invalid selected index (must be an integer)\n",
+                    optarg);
+                return EXIT_FAILURE;
+            }
+            break;
+
         case 'l':
             if (sscanf(optarg, "%u", &cmdline_overrides.conf.lines) != 1) {
                 fprintf(stderr, "%s: invalid line count\n", optarg);
@@ -1492,6 +1523,21 @@ main(int argc, char *const *argv)
             cmdline_overrides.layer_set = true;
             break;
 
+        case OPT_KEYBOARD_FOCUS:
+            if (strcasecmp(optarg, "exclusive") == 0)
+                cmdline_overrides.conf.keyboard_focus = ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE;
+            else if (strcasecmp(optarg, "on-demand") == 0)
+                cmdline_overrides.conf.keyboard_focus = ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND;
+            else {
+                fprintf(
+                    stderr,
+                    "%s: invalid keyboard-focus. Must be one of 'exclusive', 'on-demand'\n",
+                    optarg);
+                return EXIT_FAILURE;
+            }
+            cmdline_overrides.keyboard_focus_set = true;
+            break;
+
         case OPT_NO_EXIT_ON_KB_LOSS:
             cmdline_overrides.conf.exit_on_kb_focus_loss = false;
             cmdline_overrides.no_exit_on_keyboard_focus_loss_set = true;
@@ -1519,6 +1565,28 @@ main(int argc, char *const *argv)
             cmdline_overrides.dmenu_mode_set = true;
             break;
 
+        case OPT_DMENU_WITH_NTH:
+            if (sscanf(optarg, "%u", &cmdline_overrides.conf.dmenu.render_column) != 1) {
+                fprintf(
+                    stderr,
+                    "%s: invalid with-nth value (must be an integer)\n",
+                    optarg);
+                return EXIT_FAILURE;
+            }
+            cmdline_overrides.dmenu_with_nth_set = true;
+            break;
+
+        case OPT_DMENU_ACCEPT_NTH:
+                    if (sscanf(optarg, "%u", &cmdline_overrides.conf.dmenu.output_column) != 1) {
+                        fprintf(
+                            stderr,
+                            "%s: invalid accept-nth value (must be an integer)\n",
+                            optarg);
+                        return EXIT_FAILURE;
+                    }
+                    cmdline_overrides.dmenu_accept_nth_set = true;
+                    break;
+
         case OPT_LOG_LEVEL: {
             int lvl = log_level_from_string(optarg);
             if (lvl < 0) {
@@ -1624,6 +1692,11 @@ main(int argc, char *const *argv)
 
     int ret = EXIT_FAILURE;
 
+    if (select != NULL && select_idx) {
+            LOG_ERRNO("--select and --select-index cannot be used at the same time");
+            return ret;
+    }
+
     struct config conf = {0};
     bool conf_successful = config_load(&conf, config_path, NULL, check_config);
     if (!conf_successful) {
@@ -1639,7 +1712,7 @@ main(int argc, char *const *argv)
     /* Apply command line overrides */
     if (cmdline_overrides.conf.output != NULL) {
         free(conf.output);
-        conf.output = strdup(cmdline_overrides.conf.output);
+        conf.output = xstrdup(cmdline_overrides.conf.output);
     }
     if (cmdline_overrides.conf.prompt != NULL) {
         free(conf.prompt);
@@ -1666,21 +1739,21 @@ main(int argc, char *const *argv)
     }
     if (cmdline_overrides.conf.terminal != NULL) {
         free(conf.terminal);
-        conf.terminal = strdup(cmdline_overrides.conf.terminal);
+        conf.terminal = xstrdup(cmdline_overrides.conf.terminal);
     }
     if (cmdline_overrides.conf.launch_prefix != NULL) {
         free(conf.launch_prefix);
-        conf.launch_prefix = strdup(cmdline_overrides.conf.launch_prefix);
+        conf.launch_prefix = xstrdup(cmdline_overrides.conf.launch_prefix);
     }
     if (cmdline_overrides.conf.font != NULL) {
         free(conf.font);
-        conf.font = strdup(cmdline_overrides.conf.font);
+        conf.font = xstrdup(cmdline_overrides.conf.font);
     }
     if (cmdline_overrides.use_bold_set)
         conf.use_bold = cmdline_overrides.conf.use_bold;
     if (cmdline_overrides.conf.icon_theme != NULL) {
         free(conf.icon_theme);
-        conf.icon_theme = strdup(cmdline_overrides.conf.icon_theme);
+        conf.icon_theme = xstrdup(cmdline_overrides.conf.icon_theme);
     }
     if (cmdline_overrides.dpi_aware_set)
         conf.dpi_aware = cmdline_overrides.conf.dpi_aware;
@@ -1752,6 +1825,8 @@ main(int argc, char *const *argv)
         conf.letter_spacing = cmdline_overrides.conf.letter_spacing;
     if (cmdline_overrides.layer_set)
         conf.layer = cmdline_overrides.conf.layer;
+    if (cmdline_overrides.keyboard_focus_set)
+        conf.keyboard_focus = cmdline_overrides.conf.keyboard_focus;
     if (cmdline_overrides.no_exit_on_keyboard_focus_loss_set)
         conf.exit_on_kb_focus_loss = cmdline_overrides.conf.exit_on_kb_focus_loss;
     if (cmdline_overrides.dmenu_enabled_set)
@@ -1762,6 +1837,10 @@ main(int argc, char *const *argv)
         conf.dmenu.mode = cmdline_overrides.conf.dmenu.mode;
     if (cmdline_overrides.dmenu_exit_immediately_if_empty_set)
         conf.dmenu.exit_immediately_if_empty = cmdline_overrides.conf.dmenu.exit_immediately_if_empty;
+    if (cmdline_overrides.dmenu_with_nth_set)
+        conf.dmenu.render_column = cmdline_overrides.conf.dmenu.render_column;
+    if (cmdline_overrides.dmenu_accept_nth_set)
+        conf.dmenu.output_column = cmdline_overrides.conf.dmenu.output_column;
     if (cmdline_overrides.conf.list_executables_in_path)
         conf.list_executables_in_path = cmdline_overrides.conf.list_executables_in_path;
     if (cmdline_overrides.render_workers_set)
@@ -1778,7 +1857,7 @@ main(int argc, char *const *argv)
         conf.delayed_filter_limit = cmdline_overrides.conf.delayed_filter_limit;
     if (cmdline_overrides.conf.cache_path != NULL) {
         free(conf.cache_path);
-        conf.cache_path = strdup(cmdline_overrides.conf.cache_path);
+        conf.cache_path = xstrdup(cmdline_overrides.conf.cache_path);
     }
 
     if (conf.dmenu.enabled) {
@@ -1883,6 +1962,7 @@ main(int argc, char *const *argv)
         .themes = &themes,
         .icon_lock = &icon_lock,
         .select_initial = select,
+        .select_initial_idx = select_idx,
         .event_fd = -1,
         .dmenu_abort_fd = dmenu_abort_fd,
     };
