@@ -28,6 +28,8 @@
 #include "log.h"
 #include "char32.h"
 #include "wayland.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 #define min(x, y) ((x < y) ? (x) : (y))
 #define max(x, y) ((x > y) ? (x) : (y))
@@ -171,7 +173,7 @@ match_fzf(const char32_t *haystack, size_t haystack_len,
         if (pos != NULL) {
             assert(pos_count != NULL);
             (*pos_count)++;
-            *pos = reallocarray(*pos, *pos_count, sizeof((*pos)[0]));
+            *pos = xreallocarray(*pos, *pos_count, sizeof((*pos)[0]));
             (*pos)[*pos_count - 1].start = longest_match_ofs;
             (*pos)[*pos_count - 1].len = longest_match_len;
         }
@@ -228,9 +230,9 @@ match_levenshtein(struct matches *matches,
     if (src_len < pat_len)
         return NULL;
 
-    struct levenshtein_matrix **m = calloc(pat_len + 1, sizeof(m[0]));
+    struct levenshtein_matrix **m = xcalloc(pat_len + 1, sizeof(m[0]));
     for (size_t i = 0; i < pat_len + 1; i++)
-        m[i] = calloc(src_len + 1, sizeof(m[0][0]));
+        m[i] = xcalloc(src_len + 1, sizeof(m[0][0]));
 
     levenshtein_distance(src, src_len, pat, pat_len, m);
 
@@ -364,7 +366,7 @@ matches_init(struct fdm *fdm, const struct prompt *prompt,
              size_t fuzzy_max_distance, uint16_t workers,
              size_t delay_ms, size_t delay_limit)
 {
-    struct matches *matches = malloc(sizeof(*matches));
+    struct matches *matches = xmalloc(sizeof(*matches));
     *matches = (struct matches) {
         .fdm = fdm,
         .prompt = prompt,
@@ -401,20 +403,10 @@ matches_init(struct fdm *fdm, const struct prompt *prompt,
         }
 
         matches->workers.threads =
-            calloc(workers, sizeof(matches->workers.threads[0]));
-
-        if (matches->workers.threads == NULL) {
-            LOG_ERRNO("failed to allocate match worker threads array");
-            goto err_free_semaphores_and_lock;
-        }
+            xcalloc(workers, sizeof(matches->workers.threads[0]));
 
         for (size_t i = 0; i < workers; i++) {
-            struct thread_context *ctx = malloc(sizeof(*ctx));
-            if (ctx == NULL) {
-                LOG_ERRNO("failed to allocate match worker thread context");
-                goto err_free_semaphores_and_lock;
-            }
-
+            struct thread_context *ctx = xmalloc(sizeof(*ctx));
             *ctx = (struct thread_context){
                 .matches = matches,
                 .my_id = 1 + i,
@@ -529,7 +521,7 @@ matches_set_applications(struct matches *matches,
         return;
     }
 
-    matches->matches = reallocarray(
+    matches->matches = xreallocarray(
         matches->matches, applications->count, sizeof(matches->matches[0]));
 
     const size_t old_size = matches->matches_size;
@@ -608,7 +600,7 @@ size_t
 match_get_idx(const struct matches *matches, size_t idx)
 {
     const size_t page_no = matches_get_page(matches);
-    const size_t items_on_page __attribute__((unused)) = matches_get_count(matches);
+    const size_t items_on_page UNUSED = matches_get_count(matches);
 
     LOG_DBG(
         "page-count: %zu, page-no: %zu, items-on-page: %zu, idx: %zu, max: %zu, "
@@ -638,9 +630,9 @@ matches_get_match(const struct matches *matches)
 }
 
 size_t
-matches_get_application_count(const struct matches *matches)
+matches_get_application_visible_count(const struct matches *matches)
 {
-    return matches->applications->count;
+    return matches->applications->visible_count;
 }
 
 size_t
@@ -720,6 +712,16 @@ matches_selected_first(struct matches *matches)
     return true;
 }
 
+bool
+matches_selected_set(struct matches *matches, size_t idx)
+{
+    if (matches == NULL || matches->match_count <= 0 || idx >= matches->match_count)
+        return false;
+
+    matches->selected = idx;
+    return true;
+}
+
 bool
 matches_selected_last(struct matches *matches)
 {
@@ -819,6 +821,10 @@ match_compar(const void *_a, const void *_b)
         return -1;
     else if (a->application->count < b->application->count)
         return 1;
+    else if (a->pos_count > 0 && b->pos_count > 0 && a->pos[0].start < b->pos[0].start)
+        return -1;
+    else if (a->pos_count > 0 && b->pos_count > 0 && a->pos[0].start > b->pos[0].start)
+        return 1;
     else if (a->pos_count > 0 && b->pos_count > 0 &&
              a->application->title != NULL && b->application->title != NULL)
     {
@@ -905,17 +911,11 @@ match_app(struct matches *matches, struct application *app,
                     /* Extend last match position */
                     pos[pos_count - 1].len += match_len;
                 } else {
-                    size_t new_pos_count = pos_count + 1;
-                    struct match_substring *new_pos =
-                        reallocarray(pos, new_pos_count, sizeof(new_pos[0]));
+                    pos_count += 1;
+                    pos = xreallocarray(pos, pos_count, sizeof(pos[0]));
 
-                    if (new_pos != NULL) {
-                        pos = new_pos;
-                        pos_count = new_pos_count;
-
-                        pos[pos_count - 1].start = m - app->title_lowercase;
-                        pos[pos_count - 1].len = match_len;
-                    }
+                    pos[pos_count - 1].start = m - app->title_lowercase;
+                    pos[pos_count - 1].len = match_len;
                 }
             }
 
@@ -1302,7 +1302,7 @@ match_thread(void *_ctx)
     pthread_sigmask(SIG_SETMASK, &mask, NULL);
 
     char proc_title[16];
-    snprintf(proc_title, sizeof(proc_title), "fuzzel:match:%d", my_id);
+    xsnprintf(proc_title, sizeof(proc_title), "fuzzel:match:%d", my_id);
 
     if (pthread_setname_np(pthread_self(), proc_title) < 0)
         LOG_ERRNO("render worker %d: failed to set process title", my_id);
@@ -1445,18 +1445,9 @@ matches_update_internal(struct matches *matches, bool incremental)
 
     LOG_DBG("match update begin");
 
-    char32_t *copy = c32dup(ptext);
-    char32_t **tokens = malloc(sizeof(tokens[0]));
-    size_t *tok_lengths = malloc(sizeof(tok_lengths[0]));
-
-    if (copy == NULL || tokens == NULL || tok_lengths == NULL) {
-        LOG_ERR("failed to allocate tokens");
-        free(copy);
-        free(tokens);
-        free(tok_lengths);
-        goto unlock_and_return;
-    }
-
+    char32_t *copy = xc32dup(ptext);
+    char32_t **tokens = xmalloc(sizeof(tokens[0]));
+    size_t *tok_lengths = xmalloc(sizeof(tok_lengths[0]));
     size_t tok_count = 1;
     tokens[0] = copy;
     tok_lengths[0] = 0;
@@ -1475,26 +1466,11 @@ matches_update_internal(struct matches *matches, bool incremental)
         } else {
             tok_count++;
 
-            char32_t **new_tokens = reallocarray(
-                tokens, tok_count, sizeof(tokens[0]));
-            size_t *new_tok_lengths = reallocarray(
+            tokens = xreallocarray(tokens, tok_count, sizeof(tokens[0]));
+
+            tok_lengths = xreallocarray(
                 tok_lengths, tok_count, sizeof(tok_lengths[0]));
 
-            if (new_tokens == NULL || new_tok_lengths == NULL) {
-                if (new_tokens != NULL)
-                    free(new_tokens);
-                else
-                    free(tokens);
-                if (new_tok_lengths != NULL)
-                    free(new_tok_lengths);
-                else
-                    free(tok_lengths);
-                free(copy);
-                goto unlock_and_return;
-            }
-
-            tokens = new_tokens;
-            tok_lengths = new_tok_lengths;
             tokens[tok_count - 1] = p + 1;
             tok_lengths[tok_count - 1] = 0;
         }
@@ -1528,7 +1504,7 @@ matches_update_internal(struct matches *matches, bool incremental)
 
         if (incremental) {
             matches->workers.old_matches =
-                reallocarray(matches->workers.old_matches,
+                xreallocarray(matches->workers.old_matches,
                              matches->match_count,
                              sizeof(matches->workers.old_matches[0]));
 
@@ -1587,10 +1563,6 @@ matches_update_internal(struct matches *matches, bool incremental)
         matches->workers.tok_lengths = NULL;
     }
 
-    free(tok_lengths);
-    free(tokens);
-    free(copy);
-
     LOG_DBG("match update done");
 
     /* Sort */
@@ -1607,6 +1579,10 @@ matches_update_internal(struct matches *matches, bool incremental)
     if (matches->selected >= matches->match_count && matches->selected > 0)
         matches->selected = matches->match_count - 1;
 
+    free(tok_lengths);
+    free(tokens);
+    free(copy);
+
 unlock_and_return:
     matches_unlock(matches);
 }
@@ -1639,7 +1615,7 @@ matches_update(struct matches *matches)
         matches->delayed_update_type = DELAYED_FULL_UPDATE;
         if (arm_delayed_timer(matches))
             return;
-    } 
+    }
 
     matches_update_internal(matches, false);
 }
