@@ -8,7 +8,9 @@
 #include <stdlib.h>
 #include <string.h>
 #include <threads.h>
+#include <uchar.h>
 
+#include "column.h"
 #include "macros.h"
 #if HAS_INCLUDE(<pthread_np.h>)
 #include <pthread_np.h>
@@ -35,6 +37,8 @@
 #include "char32.h"
 #include "icon.h"
 #include "stride.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 #define min(x, y) ((x) < (y) ? (x) : (y))
 #define max(x, y) ((x) > (y) ? (x) : (y))
@@ -165,7 +169,7 @@ render_background(const struct render *render, struct buffer *buf)
        selection "box" from overlapping the corners */
     const unsigned int radius =
         min(render->conf->border.radius,
-            min(render->x_margin,
+            max(render->x_margin,
                 render->y_margin));
 
     if (radius == 0) {
@@ -256,9 +260,10 @@ render_baseline(const struct render *render)
      * otherwise the baseline is simply 'descent' pixels above the
      * bottom of the cell
      */
-    const int glyph_top_y = render->conf->line_height.px >= 0
-        ? round((line_height - font_height) / 2.)
-        : 0;
+    const int glyph_top_y =
+        render->conf->line_height.px >= 0 && line_height >= font_height
+            ? round((line_height - font_height) / 2.)
+            : 0;
 
     return line_height - glyph_top_y - font->descent;
 }
@@ -277,25 +282,21 @@ render_match_count(const struct render *render, struct buffer *buf,
             ? render->subpixel
             : FCFT_SUBPIXEL_NONE;
 
-    size_t total_count = matches_get_application_count(matches);
+    size_t total_count = matches_get_application_visible_count(matches);
     size_t match_count = matches_get_total_count(matches);
     const char32_t *ptext = prompt_text(prompt);
 
-    char *text = NULL;
-    int chars = asprintf(
-        &text, "%zu/%zu",
-        ptext[0] == U'\0' ? total_count : match_count, total_count);
-
-    if (chars < 0)
-        return 0;
+    char text[64];
+    size_t count = (ptext[0] == U'\0') ? total_count : match_count;
+    size_t chars = xsnprintf(text, sizeof(text), "%zu/%zu", count, total_count);
 
     /* fcft wants UTF-32. Since we only use ASCII... */
-    uint32_t *wtext = malloc(chars * sizeof(wtext[0]));
+    uint32_t wtext[64];
     for (size_t i = 0; i < chars; i++)
         wtext[i] = (uint32_t)(unsigned char)text[i];
 
     int width = -1;
-    int x = buf->width -  render->border_size - render->x_margin;
+    int x = buf->width - render->border_size - render->x_margin;
     int y = render->border_size + render->y_margin +
             (render->row_height + font->height) / 2 - font->descent;
 
@@ -319,14 +320,12 @@ render_match_count(const struct render *render, struct buffer *buf,
             }
 
             fcft_text_run_destroy(run);
-            free(text);
-            free(wtext);
             return width;
         }
     }
 
-    const struct fcft_glyph **glyphs = malloc(chars * sizeof(glyphs[0]));
-    long *x_kern = malloc(chars * sizeof(x_kern[0]));
+    const struct fcft_glyph **glyphs = xmalloc(chars * sizeof(glyphs[0]));
+    long *x_kern = xmalloc(chars * sizeof(x_kern[0]));
     char32_t prev = 0;
 
     width = 0;
@@ -363,8 +362,6 @@ render_match_count(const struct render *render, struct buffer *buf,
 
     free(x_kern);
     free(glyphs);
-    free(text);
-    free(wtext);
     return width;
 }
 
@@ -381,7 +378,7 @@ render_cursor(const struct render *render, int x, int baseline, pixman_image_t *
             PIXMAN_OP_SRC, pix, &render->pix_input_color,
             1, &(pixman_rectangle16_t){
                 x,
-                baseline - render->font->ascent,
+                baseline + render->font->descent - height,
                 font->underline.thickness,
                 height});
     } else {
@@ -613,9 +610,9 @@ render_match_text(pixman_image_t *pix, double *_x, double _y, double max_x,
         count = (*run)->count;
     } else {
         count = c32len(text);
-        glyphs = malloc(count * sizeof(glyphs[0]));
-        clusters = malloc(count * sizeof(clusters[0]));
-        kern = malloc(count * sizeof(kern[0]));
+        glyphs = xmalloc(count * sizeof(glyphs[0]));
+        clusters = xmalloc(count * sizeof(clusters[0]));
+        kern = xmalloc(count * sizeof(kern[0]));
 
         for (size_t i = 0; i < count; i++) {
             const struct fcft_glyph *glyph = fcft_rasterize_char_utf32(
@@ -765,7 +762,7 @@ render_svg_nanosvg(struct icon *icon, int x, int y, int size,
 
         float scale = svg->width > svg->height ? size / svg->width : size / svg->height;
 
-        uint8_t *data = malloc(size * size * 4);
+        uint8_t *data = xmalloc(size * size * 4);
         nsvgRasterize(rast, svg, 0, 0, scale, data, size, size, size * 4);
 
         img = pixman_image_create_bits_no_clear(
@@ -853,30 +850,41 @@ render_png_libpng(struct icon *icon, int x, int y, int size,
             &scale_transform, &_scale_transform);
         pixman_image_set_transform(png, &scale_transform);
 
-        int param_count = 0;
-        pixman_kernel_t kernel = PIXMAN_KERNEL_LANCZOS3;
-        pixman_fixed_t *params = pixman_filter_create_separable_convolution(
-            &param_count,
-            pixman_double_to_fixed(1. / scale),
-            pixman_double_to_fixed(1. / scale),
-            kernel, kernel,
-            kernel, kernel,
-            pixman_int_to_fixed(1),
-            pixman_int_to_fixed(1));
+        if (height >= 1000) {
+            if (!icon->png_size_warned) {
+                LOG_WARN(
+                    "%s: PNG is too large (%dx%d); "
+                    "downscaling using a less precise filter",
+                    icon->path, width, height);
+                icon->png_size_warned = true;
+            }
+            pixman_image_set_filter(png, PIXMAN_FILTER_FAST, NULL, 0);
+        } else {
+            int param_count = 0;
+            pixman_kernel_t kernel = PIXMAN_KERNEL_LANCZOS3;
+            pixman_fixed_t *params = pixman_filter_create_separable_convolution(
+                &param_count,
+                pixman_double_to_fixed(1. / scale),
+                pixman_double_to_fixed(1. / scale),
+                kernel, kernel,
+                kernel, kernel,
+                pixman_int_to_fixed(1),
+                pixman_int_to_fixed(1));
 
-        if (params != NULL || param_count == 0) {
-            pixman_image_set_filter(
-                png, PIXMAN_FILTER_SEPARABLE_CONVOLUTION,
-                params, param_count);
+            if (params != NULL || param_count == 0) {
+                pixman_image_set_filter(
+                    png, PIXMAN_FILTER_SEPARABLE_CONVOLUTION,
+                    params, param_count);
+            }
+
+            free(params);
         }
 
-        free(params);
-
         width *= scale;
         height *= scale;
 
         int stride = stride_for_format_and_width(fmt, width);
-        uint8_t *data = malloc(height * stride);
+        uint8_t *data = xmalloc(height * stride);
         pixman_image_t *scaled_png = pixman_image_create_bits_no_clear(
             fmt, width, height, (uint32_t *)data, stride);
         pixman_image_composite32(
@@ -1042,7 +1050,7 @@ render_one_match_entry(const struct render *render, const struct matches *matche
         char32_t *newline = c32chr(match->application->title, U'\n');
 
         if (newline != NULL) {
-            char32_t *render_title = c32dup(match->application->title);
+            char32_t *render_title = xc32dup(match->application->title);
 
             newline = render_title + (newline - match->application->title);
             *newline = U' ';
@@ -1055,6 +1063,17 @@ render_one_match_entry(const struct render *render, const struct matches *matche
             /* No newlines, use title as-is */
             match->application->render_title = match->application->title;
         }
+
+        if (render->conf->dmenu.enabled &&
+            render->conf->dmenu.render_column > 0)
+        {
+            char32_t *new_render_title =
+                nth_column(match->application->render_title, render->conf->dmenu.render_column);
+
+            if (match->application->render_title != match->application->title)
+                free(match->application->render_title);
+            match->application->render_title = new_render_title;
+        }
     }
 
     const int y = first_row + render_baseline(render) + idx * render->row_height;
@@ -1153,7 +1172,7 @@ render_worker_thread(void *_ctx)
     pthread_sigmask(SIG_SETMASK, &mask, NULL);
 
     char proc_title[16];
-    snprintf(proc_title, sizeof(proc_title), "fuzzel:rend:%d", my_id);
+    xsnprintf(proc_title, sizeof(proc_title), "fuzzel:rend:%d", my_id);
 
     if (pthread_setname_np(pthread_self(), proc_title) < 0)
         LOG_ERRNO("render worker %d: failed to set process title", my_id);
@@ -1217,7 +1236,7 @@ render_worker_thread(void *_ctx)
 struct render *
 render_init(const struct config *conf, mtx_t *icon_lock)
 {
-    struct render *render = calloc(1, sizeof(*render));
+    struct render *render = xcalloc(1, sizeof(*render));
     *render = (struct render){
         .conf = conf,
         .icon_lock = icon_lock,
@@ -1248,20 +1267,11 @@ render_init(const struct config *conf, mtx_t *icon_lock)
         goto err_free_semaphores;
     }
 
-    render->workers.threads = calloc(render->conf->render_worker_count,
-                                     sizeof(render->workers.threads[0]));
-    if (render->workers.threads == NULL) {
-        LOG_ERRNO("failed to allocate render worker threads array");
-        goto err_free_semaphores_and_lock;
-    }
+    render->workers.threads = xcalloc(render->conf->render_worker_count,
+                                      sizeof(render->workers.threads[0]));
 
     for (size_t i = 0; i < render->conf->render_worker_count; i++) {
-        struct thread_context *ctx = malloc(sizeof(*ctx));
-        if (ctx == NULL) {
-            LOG_ERRNO("failed to allocate render worker thread context");
-            goto err_free_semaphores_and_lock;
-        }
-
+        struct thread_context *ctx = xmalloc(sizeof(*ctx));
         *ctx = (struct thread_context){
             .render = render,
             .my_id = 1 + i,
