@@ -42,6 +42,8 @@
 #include "prompt.h"
 #include "render.h"
 #include "shm.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 #define min(x, y) ((x) < (y) ? (x) : (y))
 
@@ -145,6 +147,7 @@ struct wayland {
 
     struct wl_callback *frame_cb;
     bool need_refresh;
+    bool is_configured;
 
     struct wl_display *display;
     struct wl_registry *registry;
@@ -157,6 +160,8 @@ struct wayland {
     struct wl_surface *surface;
     struct zwlr_layer_shell_v1 *layer_shell;
     struct zwlr_layer_surface_v1 *layer_surface;
+    bool has_zwlr_layer_shell_kb_inter_on_demand;
+
     struct wp_viewport *viewport;
     struct wp_fractional_scale_v1 *fractional_scale;
 
@@ -488,7 +493,7 @@ token_done(void *data, struct xdg_activation_token_v1 *token,
            const char *name)
 {
     char **xdg_activation_token = data;
-    *xdg_activation_token = strdup(name);
+    *xdg_activation_token = xstrdup(name);
 }
 
 static const struct xdg_activation_token_v1_listener token_listener = {
@@ -533,7 +538,7 @@ execute_selected(struct seat *seat, bool as_is, int custom_success_exit_code)
     ssize_t index = app != NULL ? app->index : -1;
 
     if (wayl->conf->dmenu.enabled) {
-        dmenu_execute(app, index, wayl->prompt, wayl->conf->dmenu.mode);
+        dmenu_execute(app, index, wayl->prompt, wayl->conf->dmenu.mode, wayl->conf->dmenu.output_column);
         wayl->exit_code = custom_success_exit_code >= 0
             ? custom_success_exit_code
             : EXIT_SUCCESS;
@@ -651,15 +656,14 @@ execute_binding(struct seat *seat, const struct key_binding *binding, bool *refr
         if (wayl->conf->dmenu.enabled) {
             char *chars = ac32tombs(match->application->title);
             if (chars != NULL) {
-                *refresh = prompt_insert_chars(
-                    wayl->prompt, chars, strlen(chars));
+                prompt_insert_chars(wayl->prompt, chars, strlen(chars));
+                *refresh = true;
                 free(chars);
             }
         } else if (match->application->exec != NULL) {
-            *refresh = prompt_insert_chars(
-                wayl->prompt,
-                match->application->exec,
-                strlen(match->application->exec));
+            const char *chars = match->application->exec;
+            prompt_insert_chars(wayl->prompt, chars, strlen(chars));
+            *refresh = true;
         }
 
         if (*refresh)
@@ -766,82 +770,18 @@ execute_binding(struct seat *seat, const struct key_binding *binding, bool *refr
     return false;
 }
 
-static bool
-symbol_is_keypad(xkb_keysym_t sym)
-{
-    switch (sym) {
-    default:
-        return false;
-
-    case XKB_KEY_KP_Space:
-    case XKB_KEY_KP_Tab:
-    case XKB_KEY_KP_Enter:
-    case XKB_KEY_KP_F1:
-    case XKB_KEY_KP_F2:
-    case XKB_KEY_KP_F3:
-    case XKB_KEY_KP_F4:
-    case XKB_KEY_KP_Home:
-    case XKB_KEY_KP_Left:
-    case XKB_KEY_KP_Up:
-    case XKB_KEY_KP_Right:
-    case XKB_KEY_KP_Down:
-    case XKB_KEY_KP_Prior:
-    case XKB_KEY_KP_Next:
-    case XKB_KEY_KP_End:
-    case XKB_KEY_KP_Begin:
-    case XKB_KEY_KP_Insert:
-    case XKB_KEY_KP_Delete:
-    case XKB_KEY_KP_Equal:
-    case XKB_KEY_KP_Multiply:
-    case XKB_KEY_KP_Add:
-    case XKB_KEY_KP_Separator:
-    case XKB_KEY_KP_Subtract:
-    case XKB_KEY_KP_Decimal:
-    case XKB_KEY_KP_Divide:
-    case XKB_KEY_KP_0:
-    case XKB_KEY_KP_1:
-    case XKB_KEY_KP_2:
-    case XKB_KEY_KP_3:
-    case XKB_KEY_KP_4:
-    case XKB_KEY_KP_5:
-    case XKB_KEY_KP_6:
-    case XKB_KEY_KP_7:
-    case XKB_KEY_KP_8:
-    case XKB_KEY_KP_9:
-        return true;
-    }
-}
-
 static void
 keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
              uint32_t time, uint32_t key, uint32_t state)
 {
-    /* TODO: these can’t be static - need to reload on keymap changes */
-    static bool mod_masks_initialized = false;
-    static xkb_mod_mask_t shift = -1;
-    static xkb_mod_mask_t ctrl = -1;
-    static xkb_mod_mask_t alt = -1;
-    static xkb_mod_mask_t super = -1;
-
     struct seat *seat = data;
     struct wayland *wayl = seat->wayl;
 
-    if (!mod_masks_initialized) {
-        mod_masks_initialized = true;
-
-        xkb_mod_index_t shift_idx = xkb_keymap_mod_get_index(
-            seat->kbd.xkb_keymap, XKB_MOD_NAME_SHIFT);
-        xkb_mod_index_t ctrl_idx = xkb_keymap_mod_get_index(
-            seat->kbd.xkb_keymap, XKB_MOD_NAME_CTRL);
-        xkb_mod_index_t alt_idx = xkb_keymap_mod_get_index(
-            seat->kbd.xkb_keymap, XKB_MOD_NAME_ALT);
-        xkb_mod_index_t super_idx = xkb_keymap_mod_get_index(
-            seat->kbd.xkb_keymap, XKB_MOD_NAME_LOGO);
-
-        shift = shift_idx != XKB_MOD_INVALID ? 1 << shift_idx : 0;
-        ctrl = ctrl_idx != XKB_MOD_INVALID ? 1 << ctrl_idx : 0;
-        alt = alt_idx != XKB_MOD_INVALID ? 1 << alt_idx : 0;
-        super = super_idx != XKB_MOD_INVALID ? 1 << super_idx : 0;
+    if (seat->kbd.xkb == NULL ||
+        seat->kbd.xkb_keymap == NULL ||
+        seat->kbd.xkb_state == NULL)
+    {
+        return;
     }
 
     if (state == XKB_KEY_UP) {
@@ -868,16 +808,13 @@ keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
 
     xkb_mod_mask_t mods = xkb_state_serialize_mods(
         seat->kbd.xkb_state, XKB_STATE_MODS_EFFECTIVE);
-    const xkb_mod_mask_t consumed = xkb_state_key_get_consumed_mods2(
+    xkb_mod_mask_t consumed = xkb_state_key_get_consumed_mods2(
         seat->kbd.xkb_state, key, XKB_CONSUMED_MODE_XKB);
     const xkb_mod_mask_t locked = xkb_state_serialize_mods(
         seat->kbd.xkb_state, XKB_STATE_MODS_LOCKED);
 
-    const xkb_mod_mask_t significant = shift | ctrl | alt | super;
-    const xkb_mod_mask_t bind_mods
-        = mods & significant & ~locked;
-    const xkb_mod_mask_t bind_consumed =
-        consumed & significant & ~locked;
+    mods &= ~locked;
+    consumed &= ~locked;
 
     const xkb_layout_index_t layout_idx =
         xkb_state_key_get_layout(seat->kbd.xkb_state, key);
@@ -894,9 +831,8 @@ keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
     }
 #endif
 
-    LOG_DBG("sym=%u, mod=0x%08x, consumed=0x%08x, significant=0x%08x, "
-            "locked=0x%08x",
-            sym, mods, consumed, significant, locked);
+    LOG_DBG("sym=%u, mod=0x%08x, consumed=0x%08x, locked=0x%08x",
+            sym, mods, consumed, locked);
 
     const struct key_binding_set *bindings =
         key_binding_for(wayl->kb_manager, seat);
@@ -904,33 +840,12 @@ keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
 
     bool refresh = false;
     seat->kbd.serial = serial;
+
+    /* Untranslated (unshifted matching, check *exact* modifiers) */
     tll_foreach(bindings->key, it) {
         const struct key_binding *bind = &it->item;
 
-        if (bind->k.sym == sym &&
-            bind->mods == (bind_mods & ~bind_consumed) &&
-            execute_binding(seat, bind, &refresh))
-        {
-            if (refresh)
-                wayl_refresh(wayl);
-            goto maybe_repeat;
-        }
-
-        if (bind->mods != bind_mods || bind_mods != (mods & ~locked))
-            continue;
-
-        /*
-         * Skip raw key codes for keypad keys.
-         *
-         * The keypad is usually unaffected by the layout. Instead,
-         * mapping raw key codes effectively means we’ll ignore
-         * NumLock. That is, if we have a key binding for KP_PageUp,
-         * it’ll trigger regardless of NumLock setting, and thus
-         * making it impossible to use the keypad to enter numericals.
-         *
-         * See https://codeberg.org/dnkl/fuzzel/issues/192
-         */
-        if (symbol_is_keypad(bind->k.sym))
+        if (bind->mods != mods || bind->mods == 0)
             continue;
 
         for (size_t i = 0; i < raw_count; i++) {
@@ -942,6 +857,29 @@ keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
                 goto maybe_repeat;
             }
         }
+    }
+
+    /* Translated (exact matching, check *consumed* modifiers) */
+    tll_foreach(bindings->key, it) {
+        const struct key_binding *bind = &it->item;
+
+        if (bind->k.sym == sym &&
+            bind->mods == (mods & ~consumed) &&
+            execute_binding(seat, bind, &refresh))
+        {
+            if (refresh)
+                wayl_refresh(wayl);
+            goto maybe_repeat;
+        }
+    }
+
+
+    /* Raw key-code (check *exact* modifiers) */
+    tll_foreach(bindings->key, it) {
+        const struct key_binding *bind = &it->item;
+
+        if (bind->mods != mods || bind->mods == 0)
+            continue;
 
         tll_foreach(bind->k.key_codes, code) {
             if (code->item == key &&
@@ -954,7 +892,7 @@ keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
         }
     }
 
-    if ((bind_mods & ~bind_consumed) != 0)
+    if ((mods & ~consumed) != 0)
         goto maybe_repeat;
 
     /*
@@ -978,8 +916,7 @@ keyboard_key(void *data, struct wl_keyboard *wl_keyboard, uint32_t serial,
     if (count == 0)
         return;
 
-    if (!prompt_insert_chars(wayl->prompt, buf, count))
-        return;
+    prompt_insert_chars(wayl->prompt, buf, count);
 
     matches_update_incremental(wayl->matches);
     wayl_refresh(wayl);
@@ -1269,7 +1206,7 @@ seat_handle_name(void *data, struct wl_seat *wl_seat, const char *name)
 {
     struct seat *seat = data;
     free(seat->name);
-    seat->name = strdup(name);
+    seat->name = xstrdup(name);
 }
 
 static const struct wl_seat_listener seat_listener = {
@@ -1357,34 +1294,26 @@ reload_font(struct wayland *wayl, float new_dpi, float new_scale)
     wayl->font_is_sized_by_dpi = will_size_using_dpi;
 
     if (need_font_reload) {
-        char **names = malloc(wayl->font_count * sizeof(names[0]));
+        char **names = xmalloc(wayl->font_count * sizeof(names[0]));
 
         /* Apply font size, possibly scaled using the scaling factor */
         for (size_t i = 0; i < wayl->font_count; i++) {
             const struct font_spec *spec = &wayl->fonts[i];
             const bool use_px_size = spec->px_size > 0;
             const float scale = wayl->font_is_sized_by_dpi ? 1. : new_scale;
-
-            char size[64];
-            size_t size_len;
             if (use_px_size) {
-                size_len = snprintf(
-                    size, sizeof(size), ":pixelsize=%d", (int)roundf(spec->px_size * scale));
+                int px = (int)roundf(spec->px_size * scale);
+                names[i] = xasprintf("%s:pixelsize=%d", spec->pattern, px);
             } else {
-                size_len = snprintf(
-                    size, sizeof(size), ":size=%.2f", spec->pt_size * scale);
+                double pt = spec->pt_size * scale;
+                names[i] = xasprintf("%s:size=%.2f", spec->pattern, pt);
             }
-
-            size_t len = strlen(spec->pattern) + size_len + 1;
-            names[i] = malloc(len);
-
-            strcpy(names[i], spec->pattern);
-            strcat(names[i], size);
         }
 
         /* Font’s DPI */
         float dpi = will_size_using_dpi ? new_dpi : 96.;
-        char attrs[256]; snprintf(attrs, sizeof(attrs), "dpi=%.2f", dpi);
+        char attrs[256];
+        xsnprintf(attrs, sizeof(attrs), "dpi=%.2f", dpi);
 
         char bold_attrs[512];
         strcpy(bold_attrs, attrs);
@@ -1527,8 +1456,8 @@ output_geometry(void *data, struct wl_output *wl_output, int32_t x, int32_t y,
     mon->dim.mm.width = physical_width;
     mon->dim.mm.height = physical_height;
     mon->inch = sqrt(pow(mon->dim.mm.width, 2) + pow(mon->dim.mm.height, 2)) * 0.03937008;
-    mon->make = make != NULL ? strdup(make) : NULL;
-    mon->model = model != NULL ? strdup(model) : NULL;
+    mon->make = make != NULL ? xstrdup(make) : NULL;
+    mon->model = model != NULL ? xstrdup(model) : NULL;
     mon->subpixel = subpixel;
     mon->transform = transform;
 
@@ -1610,7 +1539,7 @@ xdg_output_handle_name(void *data, struct zxdg_output_v1 *xdg_output,
     struct wayland *wayl = mon->wayl;
 
     free(mon->name);
-    mon->name = name != NULL ? strdup(name) : NULL;
+    mon->name = name != NULL ? xstrdup(name) : NULL;
 
     if (wayl->conf->output != NULL &&
         mon->name != NULL &&
@@ -1626,7 +1555,7 @@ xdg_output_handle_description(void *data, struct zxdg_output_v1 *xdg_output,
 {
     struct monitor *mon = data;
     free(mon->description);
-    mon->description = description != NULL ? strdup(description) : NULL;
+    mon->description = description != NULL ? xstrdup(description) : NULL;
 }
 
 static struct zxdg_output_v1_listener xdg_output_listener = {
@@ -1791,12 +1720,26 @@ handle_global(void *data, struct wl_registry *registry,
         if (!verify_iface_version(interface, version, required))
             return;
 
+#if defined(ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND_SINCE_VERSION)
+        const uint32_t preferred = ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND_SINCE_VERSION;
+        wayl->has_zwlr_layer_shell_kb_inter_on_demand =
+            version >= ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND_SINCE_VERSION;
+#else
+        const uint32_t preferred = required;
+#endif
+
         wayl->layer_shell = wl_registry_bind(
-            wayl->registry, name, &zwlr_layer_shell_v1_interface, required);
+            wayl->registry, name, &zwlr_layer_shell_v1_interface, min(version, preferred));
+
+        if (wayl->conf->keyboard_focus == ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND &&
+            !wayl->has_zwlr_layer_shell_kb_inter_on_demand) {
+            LOG_WARN("keyboard-focus=on-demand requires zwlr_layer_shell_v1 version: %d, found version: %d",
+                ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_ON_DEMAND_SINCE_VERSION, version);
+        }
     }
 
     else if (strcmp(interface, wl_seat_interface.name) == 0) {
-        const uint32_t required = 4;
+        const uint32_t required = 5;
         if (!verify_iface_version(interface, version, required))
             return;
 
@@ -1971,6 +1914,7 @@ layer_surface_configure(void *data, struct zwlr_layer_surface_v1 *surface,
 {
     struct wayland *wayl = data;
     zwlr_layer_surface_v1_ack_configure(surface, serial);
+    wayl->is_configured = true;
     wayl_refresh(wayl);
 }
 
@@ -2076,6 +2020,9 @@ frame_callback(void *data, struct wl_callback *wl_callback, uint32_t callback_da
 void
 wayl_refresh(struct wayland *wayl)
 {
+    if (!wayl->is_configured)
+        return;
+
     if (wayl->frame_cb != NULL) {
         wayl->need_refresh = true;
         return;
@@ -2294,7 +2241,7 @@ parse_font_spec(const char *font_spec, size_t *count, struct font_spec **specs)
 {
     tll(struct font_spec) fonts = tll_init();
 
-    char *copy = strdup(font_spec);
+    char *copy = xstrdup(font_spec);
     for (char *font = strtok(copy, ",");
          font != NULL;
          font = strtok(NULL, ","))
@@ -2316,7 +2263,7 @@ parse_font_spec(const char *font_spec, size_t *count, struct font_spec **specs)
     free(copy);
 
     *count = tll_length(fonts);
-    *specs = malloc(*count * sizeof((*specs)[0]));
+    *specs = xmalloc(*count * sizeof((*specs)[0]));
 
     struct font_spec *s = *specs;
     tll_foreach(fonts, it) {
@@ -2332,7 +2279,7 @@ wayl_init(const struct config *conf, struct fdm *fdm,
           struct render *render, struct prompt *prompt,
           struct matches *matches, font_reloaded_t font_reloaded_cb, void *data)
 {
-    struct wayland *wayl = malloc(sizeof(*wayl));
+    struct wayland *wayl = xmalloc(sizeof(*wayl));
     *wayl = (struct wayland){
         .conf = conf,
         .fdm = fdm,
@@ -2451,7 +2398,10 @@ wayl_init(const struct config *conf, struct fdm *fdm,
             wayl->fractional_scale, &fractional_scale_listener, wayl);
     }
 
-    zwlr_layer_surface_v1_set_keyboard_interactivity(wayl->layer_surface, 1);
+    zwlr_layer_surface_v1_set_keyboard_interactivity(wayl->layer_surface,
+        wayl->has_zwlr_layer_shell_kb_inter_on_demand
+            ? wayl->conf->keyboard_focus
+            : ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE);
 
     zwlr_layer_surface_v1_add_listener(
         wayl->layer_surface, &layer_surface_listener, wayl);
