@@ -21,6 +21,9 @@
 #include "log.h"
 #include "char32.h"
 #include "icon.h"
+#include "macros.h"
+#include "xmalloc.h"
+#include "xsnprintf.h"
 
 typedef tll(struct application *) application_llist_t;
 
@@ -87,9 +90,11 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
                    const struct locale_variants *lc_messages,
                    application_llist_t *applications)
 {
-    FILE *f = fdopen(fd, "r");
-    if (f == NULL)
+    FILE *f = fdopen(fd, "re");
+    if (f == NULL) {
+        close(fd);
         return;
+    }
 
     bool is_desktop_entry = false;
 
@@ -162,20 +167,20 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
 
                         action->generic_name =
                             default_action->generic_name != NULL
-                            ? c32dup(default_action->generic_name) : NULL;
+                            ? xc32dup(default_action->generic_name) : NULL;
                         action->comment = default_action->comment != NULL
-                            ? c32dup(default_action->comment) : NULL;
+                            ? xc32dup(default_action->comment) : NULL;
                         action->path = default_action->path != NULL
-                            ? strdup(default_action->path) : NULL;
+                            ? xstrdup(default_action->path) : NULL;
                         action->icon = default_action->icon != NULL
-                            ? strdup(default_action->icon) : NULL;
+                            ? xstrdup(default_action->icon) : NULL;
                         action->visible = default_action->visible;
                         action->use_terminal = default_action->use_terminal;
 
                         tll_foreach(default_action->keywords, it)
-                            tll_push_back(action->keywords, c32dup(it->item));
+                            tll_push_back(action->keywords, xc32dup(it->item));
                         tll_foreach(default_action->categories, it)
-                            tll_push_back(action->categories, c32dup(it->item));
+                            tll_push_back(action->categories, xc32dup(it->item));
 
                         action_is_valid = true;
                         break;
@@ -258,13 +263,13 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
             else if (strcmp(key, "Exec") == 0) {
                 free(action->exec);
                 free(action->wexec);
-                action->exec = strdup(value);
+                action->exec = xstrdup(value);
                 action->wexec = ambstoc32(value);
             }
 
             else if (strcmp(key, "Path") == 0) {
                 free(action->path);
-                action->path = strdup(value);
+                action->path = xstrdup(value);
             }
 
             else if (strcmp(key, "GenericName") == 0) {
@@ -282,7 +287,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
 
             else if (strcmp(key, "StartupWMClass") == 0) {
                 free(action->app_id);
-                action->app_id = strdup(value);
+                action->app_id = xstrdup(value);
             }
 
             else if (strcmp(key, "Comment") == 0) {
@@ -328,7 +333,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
                      action != NULL;
                      action = strtok_r(NULL, ";", &ctx))
                 {
-                    tll_push_back(action_names, strdup(action));
+                    tll_push_back(action_names, xstrdup(action));
                 }
             }
 
@@ -337,7 +342,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
                      desktop != NULL;
                      desktop = strtok_r(NULL, ";", &ctx))
                 {
-                    tll_push_back(action->onlyshowin, strdup(desktop));
+                    tll_push_back(action->onlyshowin, xstrdup(desktop));
                 }
             }
 
@@ -346,13 +351,13 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
                      desktop != NULL;
                      desktop = strtok_r(NULL, ";", &ctx))
                 {
-                    tll_push_back(action->notshowin, strdup(desktop));
+                    tll_push_back(action->notshowin, xstrdup(desktop));
                 }
             }
 
             else if (strcmp(key, "Icon") == 0) {
                 free(action->icon);
-                action->icon = strdup(value);
+                action->icon = xstrdup(value);
             }
 
             else if (strcmp(key, "Hidden") == 0 ||
@@ -397,26 +402,14 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
             a->name = ambstoc32("<no title>");
 
         if (a->use_terminal && terminal != NULL && a->exec != NULL) {
-            char *exec_with_terminal = malloc(
-                strlen(terminal) + 1 + strlen(a->exec) + 1);
-            strcpy(exec_with_terminal, terminal);
-            strcat(exec_with_terminal, " ");
-            strcat(exec_with_terminal, a->exec);
+            char *exec_with_terminal = xstrjoin3(terminal, " ", a->exec);
             free(a->exec);
             a->exec = exec_with_terminal;
         }
 
         char32_t *title = a->name;
         if (a != default_action) {
-            size_t title_len = c32len(default_action->name) +
-                3 +  /* “ — ” */
-                c32len(a->name) +
-                1;
-            title = malloc(title_len * sizeof(char32_t));
-
-            c32cpy(title, default_action->name);
-            c32cat(title, U" — ");
-            c32cat(title, a->name);
+            title = xc32join3(default_action->name, U" — ", a->name);
             free(a->name);
         }
 
@@ -434,7 +427,7 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
             ? c32len(action->comment)
             : 0;
 
-        char32_t *title_lowercase = c32dup(title);
+        char32_t *title_lowercase = xc32dup(title);
         for (size_t i = 0; i < title_len; i++)
             title_lowercase[i] = toc32lower(title_lowercase[i]);
 
@@ -453,16 +446,16 @@ parse_desktop_file(int fd, char *id, const char32_t *file_basename_lowercase,
                 it->item[i] = toc32lower(it->item[i]);
         }
 
-        struct application *app = malloc(sizeof(*app));
+        struct application *app = xmalloc(sizeof(*app));
         *app = (struct application){
-            .id = strdup(id),
+            .id = xstrdup(id),
             .index = 0,  /* Not used in application mode */
             .path = a->path,
             .exec = a->exec,
             .app_id = a->app_id,
             .title = title,
             .title_lowercase = title_lowercase,
-            .basename = c32dup(file_basename_lowercase),
+            .basename = xc32dup(file_basename_lowercase),
             .wexec = a->wexec,
             .generic_name = a->generic_name,
             .comment = a->comment,
@@ -492,16 +485,9 @@ static char *
 new_id(const char *base_id, const char *new_part)
 {
     if (base_id == NULL)
-        return strdup(new_part);
+        return xstrdup(new_part);
 
-    size_t len = strlen(base_id) + 1 + strlen(new_part) + 1;
-    char *id = malloc(len);
-
-    strcpy(id, base_id);
-    strcat(id, "-");
-    strcat(id, new_part);
-
-    return id;
+    return xstrjoin3(base_id, "-", new_part);
 }
 
 static void
@@ -519,52 +505,35 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
     struct locale_variants lc_messages = {NULL};
 
     if (locale != NULL) {
-        char *copy = strdup(locale);
-
-        char *lang = copy;
-        char *country_start = strchr(copy, '_');
-        char *encoding_start = strchr(copy, '.');
-        char *modifier_start = strchr(copy, '@');
+        char *lang = xstrdup(locale);
+        char *country_start = strchr(lang, '_');
+        char *encoding_start = strchr(lang, '.');
+        char *modifier_start = strchr(lang, '@');
 
+        // Replace found delimiters with null terminators and skip past them
         if (country_start != NULL)
-            *country_start = '\0';
+            *country_start++ = '\0';
         if (encoding_start != NULL)
-            *encoding_start = '\0';
+            *encoding_start++ = '\0';
         if (modifier_start != NULL)
-            *modifier_start = '\0';
+            *modifier_start++ = '\0';
 
-        lc_messages.lang = copy;
-        if (country_start != NULL) {
-            if (asprintf(
-                    &lc_messages.lang_country, "%s_%s", lang, country_start + 1) < 0) {
-                LOG_WARN(
-                    "failed to construct lang_country string from %s and %s: %s",
-                    lang, country_start + 1, strerror(errno));
-            }
-        }
+        lc_messages.lang = lang;
+        if (country_start != NULL)
+            lc_messages.lang_country = xstrjoin3(lang, "_", country_start);
 
-        if (modifier_start != NULL) {
-            if (asprintf(
-                    &lc_messages.lang_modifier, "%s@%s", lang, modifier_start + 1) < 0) {
-                LOG_WARN(
-                    "failed to construct lang@modifier string from %s and %s: %s",
-                    lang, modifier_start + 1, strerror(errno));
-            }
-        }
+        if (modifier_start != NULL)
+            lc_messages.lang_modifier = xstrjoin3(lang, "@", modifier_start);
 
-        if (country_start != NULL && modifier_start != NULL) {
-            if (asprintf(
-                    &lc_messages.lang_country_modifier,
-                    "%s_%s@%s", lang, country_start + 1, modifier_start + 1) < 0) {
-                LOG_WARN(
-                    "failed to construct lang_country@modifier string from %s, %s and %s: %s",
-                    lang, country_start + 1, modifier_start + 1, strerror(errno));
-            }
-        }
+        if (country_start != NULL && modifier_start != NULL)
+            lc_messages.lang_country_modifier = xstrjoin3(
+                    lc_messages.lang_country, "@", modifier_start);
 
         LOG_DBG("lang=%s, lang_country=%s, lang@modifier=%s, lang_country@modifier=%s",
-                lc_messages.lang, lc_messages.lang_country,
-                lc_messages.lang_modifier, lc_messages.lang_country_modifier);
+                lc_messages.lang,
+                loggable_str(lc_messages.lang_country),
+                loggable_str(lc_messages.lang_modifier),
+                loggable_str(lc_messages.lang_country_modifier));
     }
 
 
@@ -574,12 +543,12 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
 
         struct stat st;
         if (fstatat(base_fd, e->d_name, &st, 0) == -1) {
-            LOG_WARN("%s: failed to stat", e->d_name);
+            LOG_WARN("%s: failed to stat: %s", e->d_name, strerror(errno));
             continue;
         }
 
         if (S_ISDIR(st.st_mode)) {
-            int dir_fd = openat(base_fd, e->d_name, O_RDONLY);
+            int dir_fd = openat(base_fd, e->d_name, O_RDONLY | O_CLOEXEC);
             if (dir_fd == -1) {
                 LOG_ERRNO("%s: failed to open", e->d_name);
                 continue;
@@ -602,9 +571,9 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                 continue;
 
             //LOG_DBG("%s", e->d_name);
-            int fd = openat(base_fd, e->d_name, O_RDONLY);
+            int fd = openat(base_fd, e->d_name, O_RDONLY | O_CLOEXEC);
             if (fd == -1)
-                LOG_WARN("%s: failed to open", e->d_name);
+                LOG_WARN("%s: failed to open: %s", e->d_name, strerror(errno));
             else {
                 const char *file_basename = strrchr(e->d_name, '/');
                 if (file_basename == NULL)
@@ -612,7 +581,7 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                 else
                     file_basename++;
 
-                const char *extension = strchr(file_basename, '.');
+                const char *extension = strrchr(file_basename, '.');
                 if (extension == NULL)
                     extension = file_basename + strlen(file_basename);
 
@@ -641,9 +610,11 @@ scan_dir(int base_fd, const char *terminal, bool include_actions,
                         fd, id, wfile_basename, terminal, include_actions,
                         filter_desktop, desktops,
                         &lc_messages, applications);
-                } else
+                    /* fd closed by parse_desktop_file() */
+                } else {
                     free(id);
-                close(fd);
+                    close(fd);
+                }
             }
         }
     }
@@ -675,7 +646,7 @@ xdg_find_programs(const char *terminal, bool include_actions,
         char path[strlen(it->item.path) + 1 + strlen("applications") + 1];
         sprintf(path, "%s/applications", it->item.path);
 
-        int fd = open(path, O_RDONLY);
+        int fd = open(path, O_RDONLY | O_CLOEXEC);
         if (fd != -1) {
             scan_dir(fd, terminal, include_actions, filter_desktop, desktops, &apps, NULL);
             close(fd);
@@ -684,11 +655,13 @@ xdg_find_programs(const char *terminal, bool include_actions,
 
     mtx_lock(&applications->lock);
     applications->count = tll_length(apps);
-    applications->v = malloc(tll_length(apps) * sizeof(applications->v[0]));
+    applications->v = xmalloc(tll_length(apps) * sizeof(applications->v[0]));
 
     size_t i = 0;
     tll_foreach(apps, it) {
         applications->v[i++] = it->item;
+        if (it->item->visible)
+            applications->visible_count++;
         tll_remove(apps, it);
     }
     tll_free(apps);
@@ -709,7 +682,7 @@ xdg_find_programs(const char *terminal, bool include_actions,
         int idx = 0;
         tll_foreach(app->keywords, it) {
             idx += swprintf(&keywords[idx],
-                            (sizeof(keywords) / sizeof(keywords[0])) - idx,
+                            ALEN(keywords) - idx,
                             L"%ls, ", (const wchar_t *)it->item);
         }
 
@@ -719,7 +692,7 @@ xdg_find_programs(const char *terminal, bool include_actions,
         idx = 0;
         tll_foreach(app->categories, it) {
             idx += swprintf(&categories[idx],
-                            (sizeof(categories) / sizeof(categories[0])) - idx,
+                            ALEN(categories) - idx,
                             L"%ls, ", (const wchar_t *)it->item);
         }
 
@@ -759,20 +732,16 @@ xdg_data_dirs(void)
     xdg_data_dirs_t ret = tll_init();
 
     const char *xdg_data_home = getenv("XDG_DATA_HOME");
+    const char *home;
     if (xdg_data_home != NULL && xdg_data_home[0] != '\0') {
-        int fd = open(xdg_data_home, O_RDONLY | O_DIRECTORY);
+        int fd = open(xdg_data_home, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
         if (fd >= 0) {
-            struct xdg_data_dir d = {.fd = fd, .path = strdup(xdg_data_home)};
+            struct xdg_data_dir d = {.fd = fd, .path = xstrdup(xdg_data_home)};
             tll_push_back(ret, d);
         }
-    } else {
-        static const char *const local = ".local/share";
-        const struct passwd *pw = getpwuid(getuid());
-
-        char *path = malloc(strlen(pw->pw_dir) + 1 + strlen(local) + 1);
-        sprintf(path, "%s/%s", pw->pw_dir, local);
-
-        int fd = open(path, O_RDONLY | O_DIRECTORY);
+    } else if ((home = getenv("HOME")) != NULL && home[0] != '\0') {
+        char *path = xstrjoin(home, "/.local/share");
+        int fd = open(path, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
         if (fd >= 0) {
             struct xdg_data_dir d = {.fd = fd, .path = path};
             tll_push_back(ret, d);
@@ -785,31 +754,31 @@ xdg_data_dirs(void)
     if (_xdg_data_dirs != NULL) {
 
         char *ctx = NULL;
-        char *copy = strdup(_xdg_data_dirs);
+        char *copy = xstrdup(_xdg_data_dirs);
 
         for (const char *tok = strtok_r(copy, ":", &ctx);
              tok != NULL;
              tok = strtok_r(NULL, ":", &ctx))
         {
-            int fd = open(tok, O_RDONLY | O_DIRECTORY);
+            int fd = open(tok, O_RDONLY | O_DIRECTORY | O_CLOEXEC);
             if (fd >= 0) {
-                struct xdg_data_dir d = {.fd = fd, .path = strdup(tok)};
+                struct xdg_data_dir d = {.fd = fd, .path = xstrdup(tok)};
                 tll_push_back(ret, d);
             }
         }
 
         free(copy);
     } else {
-        int fd1 = open("/usr/local/share", O_RDONLY | O_DIRECTORY);
-        int fd2 = open("/usr/share", O_RDONLY | O_DIRECTORY);
+        int fd1 = open("/usr/local/share", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
+        int fd2 = open("/usr/share", O_RDONLY | O_DIRECTORY | O_CLOEXEC);
 
         if (fd1 >= 0) {
-            struct xdg_data_dir d = {.fd = fd1, .path = strdup("/usr/local/share")};
+            struct xdg_data_dir d = {.fd = fd1, .path = xstrdup("/usr/local/share")};
             tll_push_back(ret, d);
         }
 
         if (fd2 >= 0) {
-            struct xdg_data_dir d = {.fd = fd2, .path = strdup("/usr/share")};
+            struct xdg_data_dir d = {.fd = fd2, .path = xstrdup("/usr/share")};
             tll_push_back(ret, d);;
         }
     }
@@ -837,7 +806,7 @@ xdg_cache_dir(void)
     const char *home = getenv("HOME");
     if (home != NULL && home[0] != '\0') {
         static char path[PATH_MAX];
-        snprintf(path, sizeof(path), "%s/.cache", home);
+        xsnprintf(path, sizeof(path), "%s/.cache", home);
         return path;
     }
 
